(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.emnapi = {}));
})(this, (function (exports) {
    var CallbackInfo = /*#__PURE__*/ (function () {
        function CallbackInfo(parent, thiz, data, args, fn) {
            this.parent = parent;
            this.thiz = thiz;
            this.data = data;
            this.args = args;
            this.fn = fn;
        }
        CallbackInfo.prototype.getNewTarget = function (envObject) {
            var thiz = this.thiz;
            if (thiz == null || thiz.constructor == null)
                return 0;
            return thiz instanceof this.fn ? envObject.ensureHandleId(thiz.constructor) : 0;
        };
        return CallbackInfo;
    }());
    var CallbackInfoStack = /*#__PURE__*/ (function () {
        function CallbackInfoStack() {
            this.current = null;
        }
        CallbackInfoStack.prototype.pop = function () {
            var current = this.current;
            if (current === null)
                return;
            this.current = current.parent;
        };
        CallbackInfoStack.prototype.push = function (thiz, data, args, fn) {
            var info = new CallbackInfo(this.current, thiz, data, args, fn);
            this.current = info;
            return info;
        };
        CallbackInfoStack.prototype.dispose = function () {
            this.current = null;
        };
        return CallbackInfoStack;
    }());

    /******************************************************************************
    Copyright (c) Microsoft Corporation.

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.
    ***************************************************************************** */
    /* global Reflect, Promise */

    var extendStatics = function(d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };

    function __extends(d, b) {
        if (typeof b !== "function" && b !== null)
            throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    }

    var Store = /*#__PURE__*/ (function () {
        function Store(capacity) {
            this._values = [undefined];
            this._values.length = capacity;
            this._size = 1;
            this._freeList = [];
        }
        Store.prototype.add = function (value) {
            var id;
            if (this._freeList.length) {
                id = this._freeList.shift();
            }
            else {
                id = this._size;
                this._size++;
                var capacity = this._values.length;
                if (id >= capacity) {
                    this._values.length = Math.ceil(capacity * 1.5);
                }
            }
            value.id = id;
            this._values[id] = value;
        };
        Store.prototype.get = function (id) {
            return this._values[id];
        };
        Store.prototype.has = function (id) {
            return this._values[id] !== undefined;
        };
        Store.prototype.remove = function (id) {
            var value = this._values[id];
            if (value) {
                value.id = 0;
                this._values[id] = undefined;
                this._freeList.push(Number(id));
            }
        };
        Store.prototype.dispose = function () {
            for (var i = 1; i < this._size; ++i) {
                var value = this._values[i];
                value === null || value === void 0 ? void 0 : value.dispose();
            }
            this._values = [undefined];
            this._size = 1;
            this._freeList = [];
        };
        return Store;
    }());

    var EnvStore = /*#__PURE__*/ (function (_super) {
        __extends(EnvStore, _super);
        function EnvStore() {
            return _super.call(this, 2) || this;
        }
        return EnvStore;
    }(Store));

    var supportNewFunction = /*#__PURE__*/ (function () {
        var f;
        try {
            f = new Function();
        }
        catch (_) {
            return false;
        }
        return typeof f === 'function';
    })();
    var _global = /*#__PURE__*/ (function () {
        if (typeof globalThis !== 'undefined')
            return globalThis;
        var g = (function () { return this; })();
        if (!g && supportNewFunction) {
            try {
                g = new Function('return this')();
            }
            catch (_) { }
        }
        if (!g) {
            if (typeof __webpack_public_path__ === 'undefined') {
                if (typeof global !== 'undefined')
                    return global;
            }
            if (typeof window !== 'undefined')
                return window;
            if (typeof self !== 'undefined')
                return self;
        }
        return g;
    })();
    var emptyException = new Error();
    var TryCatch = /*#__PURE__*/ (function () {
        function TryCatch() {
            this._exception = emptyException;
        }
        TryCatch.prototype.hasCaught = function () {
            return this._exception !== emptyException;
        };
        TryCatch.prototype.exception = function () {
            return this._exception;
        };
        TryCatch.prototype.setError = function (err) {
            this._exception = err;
        };
        TryCatch.prototype.reset = function () {
            this._exception = emptyException;
        };
        TryCatch.prototype.extractException = function () {
            var e = this._exception;
            this.reset();
            return e;
        };
        return TryCatch;
    }());
    var canSetFunctionName = /*#__PURE__*/ (function () {
        var _a;
        try {
            return Boolean((_a = Object.getOwnPropertyDescriptor(Function.prototype, 'name')) === null || _a === void 0 ? void 0 : _a.configurable);
        }
        catch (_) {
            return false;
        }
    })();
    var supportReflect = typeof Reflect === 'object';
    var supportFinalizer = (typeof FinalizationRegistry !== 'undefined') && (typeof WeakRef !== 'undefined');
    var supportWeakSymbol = /*#__PURE__*/ (function () {
        try {
            // eslint-disable-next-line symbol-description
            var sym = Symbol();
            // eslint-disable-next-line no-new
            new WeakRef(sym);
            new WeakMap().set(sym, undefined);
        }
        catch (_) {
            return false;
        }
        return true;
    })();
    var supportBigInt = typeof BigInt !== 'undefined';
    function isReferenceType(v) {
        return (typeof v === 'object' && v !== null) || typeof v === 'function';
    }
    var _setImmediate = typeof setImmediate === 'function'
        ? setImmediate
        : function (f) {
            if (typeof f !== 'function')
                return;
            var channel = new MessageChannel();
            channel.port1.onmessage = function () {
                channel.port1.onmessage = null;
                channel = undefined;
                f();
            };
            channel.port2.postMessage(null);
        };
    var _require = /*#__PURE__*/ (function () {
        var nativeRequire;
        if (typeof __webpack_public_path__ !== 'undefined') {
            nativeRequire = (function () {
                return typeof __non_webpack_require__ !== 'undefined' ? __non_webpack_require__ : undefined;
            })();
        }
        else {
            nativeRequire = (function () {
                return typeof __webpack_public_path__ !== 'undefined' ? (typeof __non_webpack_require__ !== 'undefined' ? __non_webpack_require__ : undefined) : (typeof require !== 'undefined' ? require : undefined);
            })();
        }
        return nativeRequire;
    })();
    var Buffer = /*#__PURE__*/ (function () {
        if (typeof _global.Buffer === 'function')
            return _global.Buffer;
        try {
            return _require('buffer').Buffer;
        }
        catch (_) { }
        return undefined;
    })();

    var Handle = /*#__PURE__*/ (function () {
        function Handle(id, value) {
            this.id = id;
            this.value = value;
        }
        Handle.prototype.init = function (id, value) {
            this.id = id;
            this.value = value;
        };
        Handle.prototype.data = function (envObject) {
            return envObject.getObjectBinding(this.value).data;
        };
        Handle.prototype.isEmpty = function () {
            return this.id === 0;
        };
        Handle.prototype.isNumber = function () {
            return !this.isEmpty() && typeof this.value === 'number';
        };
        Handle.prototype.isBigInt = function () {
            return !this.isEmpty() && typeof this.value === 'bigint';
        };
        Handle.prototype.isString = function () {
            return !this.isEmpty() && typeof this.value === 'string';
        };
        Handle.prototype.isFunction = function () {
            return !this.isEmpty() && typeof this.value === 'function';
        };
        Handle.prototype.isExternal = function () {
            return !this.isEmpty() && (isReferenceType(this.value) && Object.getPrototypeOf(this.value) === null);
        };
        Handle.prototype.isObject = function () {
            return !this.isEmpty() && typeof this.value === 'object' && this.value !== null;
        };
        Handle.prototype.isArray = function () {
            return !this.isEmpty() && Array.isArray(this.value);
        };
        Handle.prototype.isArrayBuffer = function () {
            return !this.isEmpty() && (this.value instanceof ArrayBuffer);
        };
        Handle.prototype.isTypedArray = function () {
            return !this.isEmpty() && (ArrayBuffer.isView(this.value)) && !(this.value instanceof DataView);
        };
        Handle.prototype.isBuffer = function () {
            return !this.isEmpty() && typeof Buffer === 'function' && Buffer.isBuffer(this.value);
        };
        Handle.prototype.isDataView = function () {
            return !this.isEmpty() && (this.value instanceof DataView);
        };
        Handle.prototype.isDate = function () {
            return !this.isEmpty() && (this.value instanceof Date);
        };
        Handle.prototype.isPromise = function () {
            return !this.isEmpty() && (this.value instanceof Promise);
        };
        Handle.prototype.isBoolean = function () {
            return !this.isEmpty() && typeof this.value === 'boolean';
        };
        Handle.prototype.isUndefined = function () {
            return !this.isEmpty() && this.value === undefined;
        };
        Handle.prototype.isSymbol = function () {
            return !this.isEmpty() && typeof this.value === 'symbol';
        };
        Handle.prototype.isNull = function () {
            return !this.isEmpty() && this.value === null;
        };
        Handle.prototype.dispose = function () {
            if (this.id === 0)
                return;
            this.id = 0;
            this.value = undefined;
        };
        return Handle;
    }());
    var ConstHandle = /*#__PURE__*/ (function (_super) {
        __extends(ConstHandle, _super);
        function ConstHandle(id, value) {
            return _super.call(this, id, value) || this;
        }
        ConstHandle.prototype.dispose = function () { };
        return ConstHandle;
    }(Handle));
    function External() {
        Object.setPrototypeOf(this, null);
    }
    External.prototype = null;
    var HandleStore = /*#__PURE__*/ (function () {
        function HandleStore() {
            this._values = [
                undefined,
                HandleStore.UNDEFINED,
                HandleStore.NULL,
                HandleStore.FALSE,
                HandleStore.TRUE,
                HandleStore.GLOBAL
            ];
            this._next = HandleStore.MIN_ID;
        }
        HandleStore.prototype.push = function (value) {
            var h;
            var next = this._next;
            var values = this._values;
            if (next < values.length) {
                h = values[next];
                h.init(next, value);
            }
            else {
                h = new Handle(next, value);
                values[next] = h;
            }
            this._next++;
            return h;
        };
        HandleStore.prototype.erase = function (start, end) {
            this._next = start;
            var values = this._values;
            for (var i = start; i < end; ++i) {
                values[i].dispose();
            }
        };
        HandleStore.prototype.get = function (id) {
            return this._values[id];
        };
        HandleStore.prototype.swap = function (a, b) {
            var values = this._values;
            var h = values[a];
            values[a] = values[b];
            values[a].id = Number(a);
            values[b] = h;
            h.id = Number(b);
        };
        HandleStore.prototype.dispose = function () {
            this._values.length = HandleStore.MIN_ID;
            this._next = HandleStore.MIN_ID;
        };
        HandleStore.UNDEFINED = new ConstHandle(1 /* GlobalHandle.UNDEFINED */, undefined);
        HandleStore.NULL = new ConstHandle(2 /* GlobalHandle.NULL */, null);
        HandleStore.FALSE = new ConstHandle(3 /* GlobalHandle.FALSE */, false);
        HandleStore.TRUE = new ConstHandle(4 /* GlobalHandle.TRUE */, true);
        HandleStore.GLOBAL = new ConstHandle(5 /* GlobalHandle.GLOBAL */, _global);
        HandleStore.MIN_ID = 6;
        return HandleStore;
    }());

    var HandleScope = /*#__PURE__*/ (function () {
        function HandleScope(handleStore, id, parentScope, start, end) {
            if (end === void 0) { end = start; }
            this.handleStore = handleStore;
            this.id = id;
            this.parent = parentScope;
            this.child = null;
            if (parentScope !== null)
                parentScope.child = this;
            this.start = start;
            this.end = end;
            this._escapeCalled = false;
        }
        HandleScope.prototype.add = function (value) {
            var h = this.handleStore.push(value);
            this.end++;
            return h;
        };
        HandleScope.prototype.addExternal = function (envObject, data) {
            var value = new External();
            var h = envObject.ctx.handleStore.push(value);
            var binding = envObject.initObjectBinding(value);
            binding.data = data;
            this.end++;
            return h;
        };
        HandleScope.prototype.dispose = function () {
            if (this.start === this.end)
                return;
            this.handleStore.erase(this.start, this.end);
        };
        HandleScope.prototype.escape = function (handle) {
            if (this._escapeCalled)
                return null;
            this._escapeCalled = true;
            if (handle < this.start || handle >= this.end) {
                return null;
            }
            this.handleStore.swap(handle, this.start);
            var h = this.handleStore.get(this.start);
            this.start++;
            this.parent.end++;
            return h;
        };
        HandleScope.prototype.escapeCalled = function () {
            return this._escapeCalled;
        };
        return HandleScope;
    }());

    var ScopeStore = /*#__PURE__*/ (function () {
        function ScopeStore() {
            this._rootScope = new HandleScope(null, 0, null, 1, HandleStore.MIN_ID);
            this.currentScope = this._rootScope;
        }
        ScopeStore.prototype.get = function (id) {
            id = Number(id);
            var scope = this.currentScope;
            while (scope !== this._rootScope) {
                if (scope.id === id) {
                    return scope;
                }
                scope = scope.parent;
            }
            return undefined;
        };
        ScopeStore.prototype.openScope = function (envObject) {
            var currentScope = this.currentScope;
            var scope = currentScope.child;
            if (scope !== null) {
                scope.start = scope.end = currentScope.end;
                scope._escapeCalled = false;
            }
            else {
                scope = new HandleScope(envObject.ctx.handleStore, currentScope.id + 1, currentScope, currentScope.end);
            }
            this.currentScope = scope;
            envObject.openHandleScopes++;
            return scope;
        };
        ScopeStore.prototype.closeScope = function (envObject) {
            if (envObject.openHandleScopes === 0)
                return;
            var scope = this.currentScope;
            this.currentScope = scope.parent;
            scope.dispose();
            envObject.openHandleScopes--;
        };
        ScopeStore.prototype.dispose = function () {
            var scope = this.currentScope;
            while (scope !== null) {
                scope.handleStore = null;
                scope.id = 0;
                scope.parent = null;
                scope.start = HandleStore.MIN_ID;
                scope.end = HandleStore.MIN_ID;
                scope._escapeCalled = false;
                var child = scope.child;
                scope.child = null;
                scope = child;
            }
            this.currentScope = null;
        };
        return ScopeStore;
    }());

    var RefStore = /*#__PURE__*/ (function (_super) {
        __extends(RefStore, _super);
        function RefStore() {
            return _super.call(this, 8) || this;
        }
        return RefStore;
    }(Store));

    var DeferredStore = /*#__PURE__*/ (function (_super) {
        __extends(DeferredStore, _super);
        function DeferredStore() {
            return _super.call(this, 8) || this;
        }
        return DeferredStore;
    }(Store));

    var RefTracker = /*#__PURE__*/ (function () {
        function RefTracker() {
            this._next = null;
            this._prev = null;
        }
        /** @virtual */
        RefTracker.prototype.finalize = function () { };
        RefTracker.prototype.link = function (list) {
            this._prev = list;
            this._next = list._next;
            if (this._next !== null) {
                this._next._prev = this;
            }
            list._next = this;
        };
        RefTracker.prototype.unlink = function () {
            if (this._prev !== null) {
                this._prev._next = this._next;
            }
            if (this._next !== null) {
                this._next._prev = this._prev;
            }
            this._prev = null;
            this._next = null;
        };
        RefTracker.finalizeAll = function (list) {
            while (list._next !== null) {
                list._next.finalize();
            }
        };
        return RefTracker;
    }());

    var Finalizer = /*#__PURE__*/ (function () {
        function Finalizer(envObject, _finalizeCallback, _finalizeData, _finalizeHint) {
            if (_finalizeCallback === void 0) { _finalizeCallback = 0; }
            if (_finalizeData === void 0) { _finalizeData = 0; }
            if (_finalizeHint === void 0) { _finalizeHint = 0; }
            this.envObject = envObject;
            this._finalizeCallback = _finalizeCallback;
            this._finalizeData = _finalizeData;
            this._finalizeHint = _finalizeHint;
        }
        Finalizer.prototype.callback = function () { return this._finalizeCallback; };
        Finalizer.prototype.data = function () { return this._finalizeData; };
        Finalizer.prototype.hint = function () { return this._finalizeHint; };
        Finalizer.prototype.resetFinalizer = function () {
            this._finalizeCallback = 0;
            this._finalizeData = 0;
            this._finalizeHint = 0;
        };
        Finalizer.prototype.dispose = function () {
            this.envObject = undefined;
        };
        return Finalizer;
    }());

    var RefBase = /*#__PURE__*/ (function (_super) {
        __extends(RefBase, _super);
        function RefBase(envObject, initial_refcount, ownership, finalize_callback, finalize_data, finalize_hint) {
            var _this = _super.call(this, envObject, finalize_callback, finalize_data, finalize_hint) || this;
            _this._next = null;
            _this._prev = null;
            _this._refcount = initial_refcount;
            _this._ownership = ownership;
            _this.link(!finalize_callback ? envObject.reflist : envObject.finalizing_reflist);
            return _this;
        }
        RefBase.finalizeAll = function (list) {
            RefTracker.finalizeAll(list);
        };
        RefBase.prototype.link = function (list) {
            RefTracker.prototype.link.call(this, list);
        };
        RefBase.prototype.unlink = function () {
            RefTracker.prototype.unlink.call(this);
        };
        RefBase.prototype.dispose = function () {
            this.unlink();
            this.envObject.dequeueFinalizer(this);
            _super.prototype.dispose.call(this);
        };
        RefBase.prototype.data = function () {
            return this._finalizeData;
        };
        RefBase.prototype.ref = function () {
            return ++this._refcount;
        };
        RefBase.prototype.unref = function () {
            if (this._refcount === 0) {
                return 0;
            }
            return --this._refcount;
        };
        RefBase.prototype.refCount = function () {
            return this._refcount;
        };
        RefBase.prototype.ownership = function () {
            return this._ownership;
        };
        RefBase.prototype.finalize = function () {
            var ownership = this._ownership;
            // Swap out the field finalize_callback so that it can not be accidentally
            // called more than once.
            var finalize_callback = this._finalizeCallback;
            var finalize_data = this._finalizeData;
            var finalize_hint = this._finalizeHint;
            this.resetFinalizer();
            this.unlink();
            var error;
            var caught = false;
            if (finalize_callback) {
                var fini = Number(finalize_callback);
                try {
                    this.envObject.callFinalizer(fini, finalize_data, finalize_hint);
                }
                catch (err) {
                    caught = true;
                    error = err;
                }
            }
            if (ownership === 0 /* Ownership.kRuntime */) {
                this.dispose();
            }
            if (caught) {
                throw error;
            }
        };
        return RefBase;
    }(Finalizer));

    var Env = /*#__PURE__*/ (function () {
        function Env(ctx, makeDynCall_vppp, makeDynCall_vp) {
            this.ctx = ctx;
            this.makeDynCall_vppp = makeDynCall_vppp;
            this.makeDynCall_vp = makeDynCall_vp;
            this.openHandleScopes = 0;
            this.instanceData = null;
            this.tryCatch = new TryCatch();
            this.refs = 1;
            this.reflist = new RefTracker();
            this.finalizing_reflist = new RefTracker();
            this.destructing = false;
            this.finalizationScheduled = false;
            this.pendingFinalizers = [];
            this.lastError = {
                errorCode: 0 /* napi_status.napi_ok */,
                engineErrorCode: 0,
                engineReserved: 0
            };
            this._bindingMap = new WeakMap();
            this.id = 0;
        }
        Env.create = function (ctx, makeDynCall_vppp, makeDynCall_vp) {
            var env = new Env(ctx, makeDynCall_vppp, makeDynCall_vp);
            ctx.envStore.add(env);
            ctx.addCleanupHook(env, function () { env.unref(); }, 0);
            return env;
        };
        Env.prototype.ref = function () {
            this.refs++;
        };
        Env.prototype.unref = function () {
            this.refs--;
            if (this.refs === 0) {
                this.dispose();
            }
        };
        Env.prototype.ensureHandle = function (value) {
            return this.ctx.ensureHandle(value);
        };
        Env.prototype.ensureHandleId = function (value) {
            return this.ensureHandle(value).id;
        };
        Env.prototype.clearLastError = function () {
            var lastError = this.lastError;
            lastError.errorCode = 0 /* napi_status.napi_ok */;
            lastError.engineErrorCode = 0;
            lastError.engineReserved = 0;
            return 0 /* napi_status.napi_ok */;
        };
        Env.prototype.setLastError = function (error_code, engine_error_code, engine_reserved) {
            if (engine_error_code === void 0) { engine_error_code = 0; }
            if (engine_reserved === void 0) { engine_reserved = 0; }
            var lastError = this.lastError;
            lastError.errorCode = error_code;
            lastError.engineErrorCode = engine_error_code;
            lastError.engineReserved = engine_reserved;
            return error_code;
        };
        Env.prototype.getReturnStatus = function () {
            return !this.tryCatch.hasCaught() ? 0 /* napi_status.napi_ok */ : this.setLastError(10 /* napi_status.napi_pending_exception */);
        };
        Env.prototype.callIntoModule = function (fn) {
            this.clearLastError();
            var r = fn(this);
            if (this.tryCatch.hasCaught()) {
                var err = this.tryCatch.extractException();
                throw err;
            }
            return r;
        };
        Env.prototype.callFinalizer = function (cb, data, hint) {
            var f = this.makeDynCall_vppp(cb);
            var env = this.id;
            var scope = this.ctx.openScope(this);
            try {
                this.callIntoModule(function () { f(env, data, hint); });
            }
            finally {
                this.ctx.closeScope(this, scope);
            }
        };
        Env.prototype.enqueueFinalizer = function (finalizer) {
            var _this = this;
            if (this.pendingFinalizers.indexOf(finalizer) === -1) {
                this.pendingFinalizers.push(finalizer);
            }
            if (!this.finalizationScheduled && !this.destructing) {
                this.finalizationScheduled = true;
                this.ref();
                _setImmediate(function () {
                    _this.finalizationScheduled = false;
                    _this.unref();
                    _this.drainFinalizerQueue();
                });
            }
        };
        Env.prototype.dequeueFinalizer = function (finalizer) {
            var index = this.pendingFinalizers.indexOf(finalizer);
            if (index !== -1) {
                this.pendingFinalizers.splice(index, 1);
            }
        };
        Env.prototype.drainFinalizerQueue = function () {
            while (this.pendingFinalizers.length > 0) {
                var refTracker = this.pendingFinalizers.shift();
                refTracker.finalize();
            }
        };
        Env.prototype.dispose = function () {
            if (this.id === 0)
                return;
            this.destructing = true;
            this.drainFinalizerQueue();
            RefBase.finalizeAll(this.finalizing_reflist);
            RefBase.finalizeAll(this.reflist);
            this.tryCatch.extractException();
            this.ctx.envStore.remove(this.id);
            this.id = 0;
        };
        Env.prototype.initObjectBinding = function (value) {
            var binding = {
                wrapped: 0,
                tag: null,
                data: 0
            };
            this._bindingMap.set(value, binding);
            return binding;
        };
        Env.prototype.getObjectBinding = function (value) {
            if (this._bindingMap.has(value)) {
                return this._bindingMap.get(value);
            }
            return this.initObjectBinding(value);
        };
        Env.prototype.setInstanceData = function (data, finalize_cb, finalize_hint) {
            if (this.instanceData) {
                this.instanceData.dispose();
            }
            this.instanceData = new RefBase(this, 0, 0 /* Ownership.kRuntime */, finalize_cb, data, finalize_hint);
        };
        Env.prototype.getInstanceData = function () {
            return this.instanceData ? this.instanceData.data() : 0;
        };
        return Env;
    }());

    var EmnapiError = /*#__PURE__*/ (function (_super) {
        __extends(EmnapiError, _super);
        function EmnapiError(message) {
            var _newTarget = this.constructor;
            var _this = _super.call(this, message) || this;
            var ErrorConstructor = _newTarget;
            var proto = ErrorConstructor.prototype;
            if (!(_this instanceof EmnapiError)) {
                var setPrototypeOf = Object.setPrototypeOf;
                if (typeof setPrototypeOf === 'function') {
                    setPrototypeOf.call(Object, _this, proto);
                }
                else {
                    // eslint-disable-next-line no-proto
                    _this.__proto__ = proto;
                }
                if (typeof Error.captureStackTrace === 'function') {
                    Error.captureStackTrace(_this, ErrorConstructor);
                }
            }
            return _this;
        }
        return EmnapiError;
    }(Error));
    Object.defineProperty(EmnapiError.prototype, 'name', {
        configurable: true,
        writable: true,
        value: 'EmnapiError'
    });
    var NotSupportWeakRefError = /*#__PURE__*/ (function (_super) {
        __extends(NotSupportWeakRefError, _super);
        function NotSupportWeakRefError(api, message) {
            return _super.call(this, "".concat(api, ": The current runtime does not support \"FinalizationRegistry\" and \"WeakRef\".").concat(message ? " ".concat(message) : '')) || this;
        }
        return NotSupportWeakRefError;
    }(EmnapiError));
    Object.defineProperty(NotSupportWeakRefError.prototype, 'name', {
        configurable: true,
        writable: true,
        value: 'NotSupportWeakRefError'
    });
    var NotSupportBigIntError = /*#__PURE__*/ (function (_super) {
        __extends(NotSupportBigIntError, _super);
        function NotSupportBigIntError(api, message) {
            return _super.call(this, "".concat(api, ": The current runtime does not support \"BigInt\".").concat(message ? " ".concat(message) : '')) || this;
        }
        return NotSupportBigIntError;
    }(EmnapiError));
    Object.defineProperty(NotSupportBigIntError.prototype, 'name', {
        configurable: true,
        writable: true,
        value: 'NotSupportBigIntError'
    });

    var StrongRef = /*#__PURE__*/ (function () {
        function StrongRef(value) {
            this._value = value;
        }
        StrongRef.prototype.deref = function () {
            return this._value;
        };
        StrongRef.prototype.dispose = function () {
            this._value = undefined;
        };
        return StrongRef;
    }());
    var Persistent = /*#__PURE__*/ (function () {
        function Persistent(value) {
            this._ref = new StrongRef(value);
        }
        Persistent.prototype.setWeak = function (param, callback) {
            if (!supportFinalizer || this._ref === undefined || this._ref instanceof WeakRef)
                return;
            var value = this._ref.deref();
            try {
                // try {
                Persistent._registry.register(value, this, this);
                var weakRef = new WeakRef(value);
                this._ref.dispose();
                this._ref = weakRef;
                // } catch (_) {
                //   Persistent._registry.register(this._ref, this, this)
                //   this._ref = new WeakRef<any>(this._ref)
                // }
                this._param = param;
                this._callback = callback;
            }
            catch (_) { }
        };
        Persistent.prototype.clearWeak = function () {
            if (!supportFinalizer || this._ref === undefined)
                return;
            if (this._ref instanceof WeakRef) {
                try {
                    Persistent._registry.unregister(this);
                }
                catch (_) { }
                this._param = undefined;
                this._callback = undefined;
                var value = this._ref.deref();
                if (value === undefined) {
                    this._ref = value;
                }
                else {
                    this._ref = new StrongRef(value);
                }
            }
        };
        Persistent.prototype.reset = function () {
            if (supportFinalizer) {
                try {
                    Persistent._registry.unregister(this);
                }
                catch (_) { }
            }
            this._param = undefined;
            this._callback = undefined;
            if (this._ref instanceof StrongRef) {
                this._ref.dispose();
            }
            this._ref = undefined;
        };
        Persistent.prototype.isEmpty = function () {
            return this._ref === undefined;
        };
        Persistent.prototype.deref = function () {
            if (this._ref === undefined)
                return undefined;
            return this._ref.deref();
        };
        Persistent._registry = supportFinalizer
            ? new FinalizationRegistry(function (value) {
                value._ref = undefined;
                var callback = value._callback;
                var param = value._param;
                value._callback = undefined;
                value._param = undefined;
                if (typeof callback === 'function') {
                    callback(param);
                }
            })
            : undefined;
        return Persistent;
    }());

    function weakCallback(ref) {
        ref.persistent.reset();
        ref.envObject.enqueueFinalizer(ref);
    }
    var Reference = /*#__PURE__*/ (function (_super) {
        __extends(Reference, _super);
        function Reference(envObject, initialRefcount, ownership, finalize_callback, finalize_data, finalize_hint) {
            if (finalize_callback === void 0) { finalize_callback = 0; }
            if (finalize_data === void 0) { finalize_data = 0; }
            if (finalize_hint === void 0) { finalize_hint = 0; }
            var _this = _super.call(this, envObject, initialRefcount >>> 0, ownership, finalize_callback, finalize_data, finalize_hint) || this;
            _this.id = 0;
            return _this;
        }
        Reference.create = function (envObject, handle_id, initialRefcount, ownership, finalize_callback, finalize_data, finalize_hint) {
            if (finalize_callback === void 0) { finalize_callback = 0; }
            if (finalize_data === void 0) { finalize_data = 0; }
            if (finalize_hint === void 0) { finalize_hint = 0; }
            var handle = envObject.ctx.handleStore.get(handle_id);
            var ref = new Reference(envObject, initialRefcount, ownership, finalize_callback, finalize_data, finalize_hint);
            envObject.ctx.refStore.add(ref);
            ref.persistent = new Persistent(handle.value);
            if (initialRefcount === 0) {
                ref._setWeak();
            }
            return ref;
        };
        Reference.prototype.ref = function () {
            if (this.persistent.isEmpty()) {
                return 0;
            }
            var count = _super.prototype.ref.call(this);
            if (count === 1) {
                var obj = this.persistent.deref();
                if (obj) {
                    this.persistent.clearWeak();
                }
            }
            return count;
        };
        Reference.prototype.unref = function () {
            if (this.persistent.isEmpty()) {
                return 0;
            }
            var oldRefcount = this.refCount();
            var refcount = _super.prototype.unref.call(this);
            if (oldRefcount === 1 && refcount === 0) {
                var obj = this.persistent.deref();
                if (obj) {
                    this._setWeak();
                }
            }
            return refcount;
        };
        Reference.prototype.get = function () {
            var obj = this.persistent.deref();
            if (obj) {
                var handle = this.envObject.ensureHandle(obj);
                return handle.id;
            }
            return 0;
        };
        Reference.prototype._setWeak = function () {
            this.persistent.setWeak(this, weakCallback);
        };
        Reference.prototype.finalize = function () {
            this.persistent.reset();
            _super.prototype.finalize.call(this);
        };
        Reference.prototype.dispose = function () {
            if (this.id === 0)
                return;
            this.persistent.reset();
            this.envObject.ctx.refStore.remove(this.id);
            _super.prototype.dispose.call(this);
            this.id = 0;
        };
        return Reference;
    }(RefBase));

    var Deferred = /*#__PURE__*/ (function () {
        function Deferred(ctx, value) {
            this.id = 0;
            this.ctx = ctx;
            this.value = value;
        }
        Deferred.create = function (ctx, value) {
            var deferred = new Deferred(ctx, value);
            ctx.deferredStore.add(deferred);
            return deferred;
        };
        Deferred.prototype.resolve = function (value) {
            this.value.resolve(value);
            this.dispose();
        };
        Deferred.prototype.reject = function (reason) {
            this.value.reject(reason);
            this.dispose();
        };
        Deferred.prototype.dispose = function () {
            this.ctx.deferredStore.remove(this.id);
            this.id = 0;
            this.value = null;
            this.ctx = null;
        };
        return Deferred;
    }());

    var CleanupHookCallback = /*#__PURE__*/ (function () {
        function CleanupHookCallback(envObject, fn, arg, order) {
            this.envObject = envObject;
            this.fn = fn;
            this.arg = arg;
            this.order = order;
        }
        return CleanupHookCallback;
    }());
    var CleanupQueue = /*#__PURE__*/ (function () {
        function CleanupQueue() {
            this._cleanupHooks = [];
            this._cleanupHookCounter = 0;
        }
        CleanupQueue.prototype.empty = function () {
            return this._cleanupHooks.length === 0;
        };
        CleanupQueue.prototype.add = function (envObject, fn, arg) {
            if (this._cleanupHooks.filter(function (hook) { return (hook.envObject === envObject && hook.fn === fn && hook.arg === arg); }).length > 0) {
                throw new Error('Can not add same fn and arg twice');
            }
            this._cleanupHooks.push(new CleanupHookCallback(envObject, fn, arg, this._cleanupHookCounter++));
        };
        CleanupQueue.prototype.remove = function (envObject, fn, arg) {
            for (var i = 0; i < this._cleanupHooks.length; ++i) {
                var hook = this._cleanupHooks[i];
                if (hook.envObject === envObject && hook.fn === fn && hook.arg === arg) {
                    this._cleanupHooks.splice(i, 1);
                    return;
                }
            }
        };
        CleanupQueue.prototype.drain = function () {
            var hooks = this._cleanupHooks.slice();
            hooks.sort(function (a, b) { return (b.order - a.order); });
            for (var i = 0; i < hooks.length; ++i) {
                var cb = hooks[i];
                if (typeof cb.fn === 'number') {
                    cb.envObject.makeDynCall_vp(cb.fn)(cb.arg);
                }
                else {
                    cb.fn(cb.arg);
                }
                this._cleanupHooks.splice(this._cleanupHooks.indexOf(cb), 1);
            }
        };
        CleanupQueue.prototype.dispose = function () {
            this._cleanupHooks.length = 0;
            this._cleanupHookCounter = 0;
        };
        return CleanupQueue;
    }());
    var NodejsWaitingRequestCounter = /*#__PURE__*/ (function () {
        function NodejsWaitingRequestCounter() {
            this.refHandle = new MessageChannel().port1;
            this.count = 0;
        }
        NodejsWaitingRequestCounter.prototype.increase = function () {
            if (this.count === 0) {
                this.refHandle.ref();
            }
            this.count++;
        };
        NodejsWaitingRequestCounter.prototype.decrease = function () {
            if (this.count === 0)
                return;
            if (this.count === 1) {
                this.refHandle.unref();
            }
            this.count--;
        };
        return NodejsWaitingRequestCounter;
    }());
    var Context = /*#__PURE__*/ (function () {
        function Context() {
            var _this = this;
            this.envStore = new EnvStore();
            this.scopeStore = new ScopeStore();
            this.refStore = new RefStore();
            this.deferredStore = new DeferredStore();
            this.handleStore = new HandleStore();
            this.cbinfoStack = new CallbackInfoStack();
            this.feature = {
                supportReflect: supportReflect,
                supportFinalizer: supportFinalizer,
                supportWeakSymbol: supportWeakSymbol,
                supportBigInt: supportBigInt,
                supportNewFunction: supportNewFunction,
                canSetFunctionName: canSetFunctionName,
                setImmediate: _setImmediate,
                Buffer: Buffer
            };
            this.cleanupQueue = new CleanupQueue();
            if (typeof process === 'object' && process !== null && typeof process.once === 'function') {
                this.refCounter = new NodejsWaitingRequestCounter();
                process.once('beforeExit', function () {
                    _this.runCleanup();
                });
            }
        }
        Context.prototype.createNotSupportWeakRefError = function (api, message) {
            return new NotSupportWeakRefError(api, message);
        };
        Context.prototype.createNotSupportBigIntError = function (api, message) {
            return new NotSupportBigIntError(api, message);
        };
        Context.prototype.createReference = function (envObject, handle_id, initialRefcount, ownership, finalize_callback, finalize_data, finalize_hint) {
            if (finalize_callback === void 0) { finalize_callback = 0; }
            if (finalize_data === void 0) { finalize_data = 0; }
            if (finalize_hint === void 0) { finalize_hint = 0; }
            return Reference.create(envObject, handle_id, initialRefcount, ownership, finalize_callback, finalize_data, finalize_hint);
        };
        Context.prototype.createDeferred = function (value) {
            return Deferred.create(this, value);
        };
        Context.prototype.createEnv = function (makeDynCall_vppp, makeDynCall_vp) {
            return Env.create(this, makeDynCall_vppp, makeDynCall_vp);
        };
        Context.prototype.getCurrentScope = function () {
            return this.scopeStore.currentScope;
        };
        Context.prototype.addToCurrentScope = function (value) {
            return this.scopeStore.currentScope.add(value);
        };
        Context.prototype.openScope = function (envObject) {
            return this.scopeStore.openScope(envObject);
        };
        Context.prototype.closeScope = function (envObject, _scope) {
            this.scopeStore.closeScope(envObject);
        };
        Context.prototype.ensureHandle = function (value) {
            switch (value) {
                case undefined: return HandleStore.UNDEFINED;
                case null: return HandleStore.NULL;
                case true: return HandleStore.TRUE;
                case false: return HandleStore.FALSE;
                case _global: return HandleStore.GLOBAL;
            }
            return this.addToCurrentScope(value);
        };
        Context.prototype.addCleanupHook = function (envObject, fn, arg) {
            this.cleanupQueue.add(envObject, fn, arg);
        };
        Context.prototype.removeCleanupHook = function (envObject, fn, arg) {
            this.cleanupQueue.remove(envObject, fn, arg);
        };
        Context.prototype.runCleanup = function () {
            while (!this.cleanupQueue.empty()) {
                this.cleanupQueue.drain();
            }
        };
        Context.prototype.increaseWaitingRequestCounter = function () {
            var _a;
            (_a = this.refCounter) === null || _a === void 0 ? void 0 : _a.increase();
        };
        Context.prototype.decreaseWaitingRequestCounter = function () {
            var _a;
            (_a = this.refCounter) === null || _a === void 0 ? void 0 : _a.decrease();
        };
        Context.prototype.dispose = function () {
            this.runCleanup();
        };
        return Context;
    }());
    var defaultContext;
    function createContext() {
        return new Context();
    }
    function getDefaultContext() {
        if (!defaultContext) {
            defaultContext = createContext();
        }
        return defaultContext;
    }

    var version = "0.36.2";

    exports.CallbackInfo = CallbackInfo;
    exports.CallbackInfoStack = CallbackInfoStack;
    exports.ConstHandle = ConstHandle;
    exports.Context = Context;
    exports.Deferred = Deferred;
    exports.DeferredStore = DeferredStore;
    exports.EmnapiError = EmnapiError;
    exports.Env = Env;
    exports.EnvStore = EnvStore;
    exports.Finalizer = Finalizer;
    exports.Handle = Handle;
    exports.HandleScope = HandleScope;
    exports.HandleStore = HandleStore;
    exports.NotSupportBigIntError = NotSupportBigIntError;
    exports.NotSupportWeakRefError = NotSupportWeakRefError;
    exports.Persistent = Persistent;
    exports.RefBase = RefBase;
    exports.RefStore = RefStore;
    exports.RefTracker = RefTracker;
    exports.Reference = Reference;
    exports.ScopeStore = ScopeStore;
    exports.Store = Store;
    exports.TryCatch = TryCatch;
    exports.createContext = createContext;
    exports.getDefaultContext = getDefaultContext;
    exports.isReferenceType = isReferenceType;
    exports.version = version;

    Object.defineProperty(exports, '__esModule', { value: true });

}));
